package tjm.tdg;

import java.util.ArrayList;
import java.util.List;

import tjm.tdg.algorithms.AStar;
import tjm.tdg.entities.Entity;
import tjm.tdg.entities.EntityType;
import tjm.tdg.entities.Team;
import tjm.tdg.entities.mobs.Berserker;
import tjm.tdg.entities.mobs.Kamikaze;
import tjm.tdg.entities.mobs.Mob;
import tjm.tdg.entities.mobs.Tactical;
import tjm.tdg.entities.towers.Tower;
import tjm.tdg.entities.towers.TowerBase;
import tjm.tdg.entities.towers.TowerBasic;
import tjm.tdg.entities.towers.TowerLongRange;
import tjm.tdg.entities.towers.TowerSplash;
import tjm.tdg.tiles.Tile;
import tjm.tdg.utils.Vector2f;
import tjm.tdg.utils.Vector2i;
import android.view.MotionEvent;

import com.threed.jpct.Interact2D;
import com.threed.jpct.Matrix;
import com.threed.jpct.Object3D;
import com.threed.jpct.SimpleVector;

public class GameWorld {
	public final Renderer renderer;
	private AStar astar = null;
	private boolean hasAddedBase = false;
	private Vector2f basePoint;
	private TowerBase baseTower;
	private final List<Vector2f> baseAllPoints = new ArrayList<Vector2f>(4);
	private int ticksSpawn = (int) 10e3;
	private int ticksTotal = 0;
	public int touchAction = 0;
	public boolean placingTower = false;
	public Object3D placingTowerModel;
	public EntityType placingTowerType;
	public final Vector2i placingTowerPos = new Vector2i();
	private final SimpleVector pickingVector = new SimpleVector();
	private int width = 0, height = 0;
	private Tile[][] tiles = null;

	public int coins = 10;
	public int deadAlly = 0;
	public int deadEnemy = 0;

	private int entitiesTowers = 0;
	private int entitiesEnemy = 0;

	public List<Entity> entities = new ArrayList<Entity>(50);

	public GameWorld(Renderer gameRenderer, int levelWidth, int levelHeight) {
		this.renderer = gameRenderer;
		this.width = levelWidth;
		this.height = levelHeight;
		this.tiles = new Tile[this.height][this.width];
		for (int y = 0; y < this.height; y++) {
			for (int x = 0; x < this.width; x++) {
				this.tiles[y][x] = new Tile(new Vector2i(x, y));
			}
		}

		this.astar = new AStar(this.width, this.height, true, false);
	}

	private void updateEntityPaths() {
		for (int i = 0; i < this.entities.size(); i++) {
			Entity entity = this.entities.get(i);
			if (entity.type.isMob()) {
				entity.waypoints = this.astar.search(entity.position, this.baseAllPoints);
			}
		}
	}

	public boolean addTower(Vector2i pos, EntityType type) {
		if (pos.x < 0 || pos.y < 0 || pos.x >= this.tiles[0].length || pos.y >= this.tiles.length) {
			return false;
		}

		if (this.entitiesTowers > Settings.MAX_ENTITIES_TOWER) {
			return false;
		}

		Tile tile = this.tiles[pos.y][pos.x];

		if (tile.isBlocked) {
			return false;
		}

		Tower tower;

		switch (type) {
		case TOWER_BASIC:
			tower = new TowerBasic(this, new Vector2f(pos.x, pos.y));
			this.coins -= Settings.TOWER_COST_BASIC;
			break;
		case TOWER_LONG_RANGE:
			tower = new TowerLongRange(this, new Vector2f(pos.x, pos.y));
			this.coins -= Settings.TOWER_COST_LONG_RANGE;
			break;
		case TOWER_SPLASH:
			tower = new TowerSplash(this, new Vector2f(pos.x, pos.y));
			this.coins -= Settings.TOWER_COST_SPLASH;
			break;
		default:
			return false;
		}

		tile.isBlocked = true;
		this.astar.setWall(tile.position.x, tile.position.y);
		updateEntityPaths();

		addEntity(tower);

		this.entitiesTowers++;

		return true;
	}

	public boolean addMob(Vector2f pos, EntityType type) {
		Mob mob;

		switch (type) {
		case KAMIKAZE:
			mob = new Kamikaze(this, pos);
			break;
		case BERSERKER:
			mob = new Berserker(this, pos);
			break;
		case TACTICAL:
			mob = new Tactical(this, pos);
			break;
		default:
			return false;
		}

		mob.waypoints = this.astar.search(pos, this.baseAllPoints);

		addEntity(mob);

		this.entitiesEnemy++;

		return true;
	}

	public void addEntity(Entity entity) {
		entity.model.translate(entity.position.x, entity.modelOffsetY, entity.position.y);

		this.entities.add(entity);
		this.renderer.world.addObject(entity.model);
	}

	public void removeEntity(Entity entity) {
		this.entities.remove(entity);
		this.renderer.world.removeObject(entity.model);
	}

	public void update(int ticks) {
		if (!this.hasAddedBase) {
			this.hasAddedBase = true;

			this.basePoint = new Vector2f(this.width / 2.0f + 0.5f, this.height / 2.0f + 0.5f);
			this.baseTower = new TowerBase(this, this.basePoint);

			this.baseAllPoints.add(this.basePoint);

			addEntity(this.baseTower);
		}

		// clean up dead entities
		for (int i = 0; i < this.entities.size(); i++) {
			Entity entity = this.entities.get(i);
			if (entity.isDead()) {
				if (entity.type.isTower()) {
					Tile tile = this.tiles[(int) entity.position.y][(int) entity.position.x];
					if (tile != null) {
						tile.isBlocked = false;
						this.astar.setClear(tile.position.x, tile.position.y);
						updateEntityPaths();
					}
					this.entitiesTowers--;
				} else if (entity.type.isMob()) {
					this.coins += ((Mob) entity).coins;
					this.entitiesEnemy--;
				}

				if (!entity.type.isProjectile()) {
					if (entity.team == Team.ALLY) {
						this.deadAlly++;
					} else if (entity.team == Team.ENEMY) {
						this.deadEnemy++;
					}
				}

				removeEntity(entity);
			}
		}

		// spawn new entities
		this.ticksSpawn += ticks;
		this.ticksTotal += ticks;
		if (this.ticksSpawn > (160 - this.ticksTotal / 100) && this.entitiesEnemy < Settings.MAX_ENTITIES_ENEMY && Math.random() < 0.4) {
			int x, y;

			this.ticksSpawn = 0;

			x = (int) (Math.random() * Settings.LEVEL_WIDTH);
			y = (int) (Math.random() * Settings.LEVEL_HEIGHT);

			if (Math.random() < 0.5) {
				if (x < Settings.LEVEL_WIDTH / 2.0) {
					x = -1;
				} else {
					x = Settings.LEVEL_WIDTH;
				}
			} else {
				if (y < Settings.LEVEL_HEIGHT / 2.0) {
					y = -1;
				} else {
					y = Settings.LEVEL_HEIGHT;
				}
			}

			EntityType type;
			int r = (int) (Math.random() * 100);
			if (r <= 45) {
				type = EntityType.BERSERKER;
			} else if (r <= 80) {
				type = EntityType.TACTICAL;
			} else {
				type = EntityType.KAMIKAZE;
			}

			addMob(new Vector2f(x, y), type);
		}

		// update entities
		for (int i = 0; i < this.entities.size(); i++) {
			this.entities.get(i).update(ticks);
		}
	}

	public void onTouchEvent(MotionEvent me) {
		if (me.getAction() == MotionEvent.ACTION_MOVE) {
			switch (this.touchAction) {
			case 1:
				Interact2D.reproject2D3D(this.renderer.world.getCamera(), this.renderer.fb, (int) (this.renderer.width - me.getX()), (int) me.getY(), Settings.CAM_OFFSET_Y, this.pickingVector);

				this.placingTowerPos.set((int) (this.pickingVector.x + Settings.CAM_OFFSET_X), (int) (this.pickingVector.y + Settings.CAM_OFFSET_Z + Settings.PLACEMENT_OFFSET));
				this.placingTowerModel.setTranslationMatrix(new Matrix());
				this.placingTowerModel.translate(this.placingTowerPos.x, -1.0f, this.placingTowerPos.y);
				break;
			}
		} else if (me.getAction() == MotionEvent.ACTION_UP) {
			switch (this.touchAction) {
			case 1:
				this.renderer.world.removeObject(this.placingTowerModel);
				addTower(this.placingTowerPos, this.placingTowerType);
				this.placingTower = false;
				this.placingTowerModel = null;
				break;
			}

			this.touchAction = 0;
		}
	}

	public boolean isGameOver() {
		if (this.baseTower == null) {
			return false;
		}

		return this.baseTower.isDead();
	}

	public String status() {
		if (this.baseTower == null) {
			return "GAME OVER";
		}

		return "HP: " + this.baseTower.health + "/" + this.baseTower.maxHealth;
	}
}
