package tjm.tdg.algorithms;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import tjm.tdg.utils.Vector2f;

class Node {
	enum Type {
		None, Clear, Wall
	}

	public Node from;
	Vector2f pos;
	public Type type;
	public double gscore;
	public double hscore;
	public double fscore;

	public Node(Node prev, Vector2f pos, Type type, double gscore, double hscore) {
		this.from = prev;
		this.pos = pos;
		this.type = type;
		this.gscore = gscore;
		this.hscore = hscore;
		this.fscore = gscore + hscore;
	}
}

public class AStar {
	private final int width, height;
	private final Node[][] grid;
	private final boolean diag;
	private final boolean wdiag;

	protected void setTile(int x, int y, Node.Type type) {
		if (!inside(x, y)) {
			return;
		}
		this.grid[x][y] = new Node(null, new Vector2f(x, y), type, 0, 0);
	}

	protected Node findNearestNode(Vector2f point, List<Node> nodes) {
		if (nodes.isEmpty()) {
			return null;
		}

		Node nearest = nodes.get(0);
		double min = point.clone().sub(nearest.pos).length();
		double len;
		for (int i = 1; i < nodes.size(); i++) {
			len = point.clone().sub(nodes.get(i).pos).length();
			if (len < min) {
				nearest = nodes.get(i);
				min = len;
			}
		}
		return nearest;
	}

	protected Node.Type neighborType(int x, int y, Node.Type def) {
		if (!inside(x, y)) {
			return def;
		}
		return this.grid[x][y].type;
	}

	protected Node.Type neighborType(int x, int y) {
		return neighborType(x, y, Node.Type.None);
	}

	protected boolean neighborClear(int x, int y, List<Node> goals) {
		return neighborType(x, y) == Node.Type.Clear || (inside(x, y) && isGoal(this.grid[x][y], goals));
	}

	protected List<Node> neighborNodes(Node curr, List<Node> goals) {
		int x = (int) curr.pos.x;
		int y = (int) curr.pos.y;
		assert inside(x, y);

		List<Node> neigh = new ArrayList<Node>(8);

		if (neighborClear(x - 1, y, goals)) {
			neigh.add(this.grid[x - 1][y]);
		}
		if (neighborClear(x + 1, y, goals)) {
			neigh.add(this.grid[x + 1][y]);
		}
		if (neighborClear(x, y - 1, goals)) {
			neigh.add(this.grid[x][y - 1]);
		}
		if (neighborClear(x, y + 1, goals)) {
			neigh.add(this.grid[x][y + 1]);
		}

		if (this.diag) {
			if (neighborClear(x - 1, y - 1, goals) && (this.wdiag || (neighborClear(x - 1, y, goals) && neighborClear(x, y - 1, goals)))) {
				neigh.add(this.grid[x - 1][y - 1]);
			}
			if (neighborClear(x + 1, y - 1, goals) && (this.wdiag || (neighborClear(x + 1, y, goals) && neighborClear(x, y - 1, goals)))) {
				neigh.add(this.grid[x + 1][y - 1]);
			}
			if (neighborClear(x - 1, y + 1, goals) && (this.wdiag || (neighborClear(x - 1, y, goals) && neighborClear(x, y + 1, goals)))) {
				neigh.add(this.grid[x - 1][y + 1]);
			}
			if (neighborClear(x + 1, y + 1, goals) && (this.wdiag || (neighborClear(x + 1, y, goals) && neighborClear(x, y + 1, goals)))) {
				neigh.add(this.grid[x + 1][y + 1]);
			}
		}

		return neigh;
	}

	protected List<Vector2f> reconstructPath(Node node) {
		List<Vector2f> list = new ArrayList<Vector2f>();

		while (node != null) {
			list.add(node.pos);
			node = node.from;
		}

		Collections.reverse(list);
		return list;
	}

	protected int insertFScore(List<Node> list, Node node) {
		for (int i = list.size() - 1; i >= 0; i--) {
			if (list.get(i).fscore <= node.fscore) {
				if (i == list.size() - 1) {
					list.add(node);
				} else {
					list.add(i + 1, node);
				}
				return i;
			}
		}
		list.add(0, node);
		return 0;
	}

	protected Vector2f nearestInside(Vector2f point) {
		Vector2f ret = point.clone();
		if (point.x < 0) {
			ret.x = 0;
		} else if (point.x >= this.width) {
			ret.x = this.width - 1;
		}
		if (point.y < 0) {
			ret.y = 0;
		} else if (point.y >= this.height) {
			ret.y = this.height - 1;
		}
		return ret;
	}

	protected void prepareForSearch() {
		int x, y;
		for (x = 0; x < this.width; x++) {
			for (y = 0; y < this.height; y++) {
				this.grid[x][y] = new Node(null, new Vector2f(x, y), this.grid[x][y].type, 0, 0);
			}
		}
	}

	public AStar(int width, int height, boolean diag, boolean wdiag) {
		int x, y;

		this.width = width;
		this.height = height;
		this.grid = new Node[width][height];
		for (x = 0; x < width; x++) {
			for (y = 0; y < height; y++) {
				this.grid[x][y] = new Node(null, new Vector2f(x, y), Node.Type.Clear, 0, 0);
			}
		}
		this.diag = diag;
		this.wdiag = wdiag;
	}

	public AStar(int width, int height, boolean diag) {
		this(width, height, diag, false);
	}

	public AStar(int width, int height) {
		this(width, height, true);
	}

	public void setWall(int x, int y) {
		setTile(x, y, Node.Type.Wall);
	}

	public void setWall(Vector2f point) {
		setWall((int) point.x, (int) point.y);
	}

	public void setClear(int x, int y) {
		setTile(x, y, Node.Type.Clear);
	}

	public void setClear(Vector2f point) {
		setWall((int) point.x, (int) point.y);
	}

	public boolean inside(int x, int y) {
		return !(x < 0 || y < 0 || x >= this.width || y >= this.height);
	}

	public boolean inside(Vector2f point) {
		return inside((int) point.x, (int) point.y);
	}

	public List<Vector2f> search(Vector2f start, List<Vector2f> pgoals) {
		prepareForSearch();

		start = nearestInside(start);
		assert inside(start);
		List<Node> goals = new ArrayList<Node>(pgoals.size());
		int i;
		for (i = 0; i < pgoals.size(); i++) {
			if (inside(pgoals.get(i))) {
				goals.add(this.grid[(int) pgoals.get(i).x][(int) pgoals.get(i).y]);
			}
		}
		if (goals.isEmpty()) {
			return null;
		}

		List<Node> open = new ArrayList<Node>();
		List<Node> closed = new ArrayList<Node>();

		Node curr = this.grid[(int) start.x][(int) start.y];
		curr.from = null;
		curr.gscore = 0;
		curr.hscore = hScore(curr, goals);
		curr.fscore = curr.gscore + curr.hscore;
		open.add(curr);

		double tgscore; // tentative g score
		boolean contains;

		while (!open.isEmpty()) {
			curr = open.get(0);
			if (isGoal(curr, goals)) {
				return reconstructPath(curr);
			}
			open.remove(0);
			closed.add(curr);
			for (Node neigh : neighborNodes(curr, goals)) {
				if (closed.contains(neigh)) {
					continue;
				}
				tgscore = curr.gscore + gScore(curr, neigh);

				contains = open.contains(neigh);
				if (!contains || tgscore < neigh.gscore) {
					neigh.from = curr;
					neigh.gscore = tgscore;
					neigh.hscore = hScore(neigh, goals);
					neigh.fscore = neigh.gscore + neigh.hscore;
					if (!contains) {
						insertFScore(open, neigh);
					}
				}
			}
		}

		return reconstructPath(curr);
	}

	public List<Vector2f> search(Vector2f start, Vector2f goal) {
		List<Vector2f> goals = new ArrayList<Vector2f>(1);
		goals.add(goal);
		return search(start, goals);
	}

	public int getWidth() {
		return this.width;
	}

	public int getHeight() {
		return this.height;
	}

	public boolean isGoal(Node curr, List<Node> goals) {
		return goals.contains(curr);
	}

	public double gScore(Node prev, Node curr) {
		return curr.pos.clone().sub(prev.pos).length();
	}

	public double hScore(Node curr, List<Node> goals) {
		Node nearest = findNearestNode(curr.pos, goals);
		if (nearest == null) {
			return 0;
		}
		return curr.pos.clone().sub(nearest.pos).length();
	}
}
