package tjm.tdg.algorithms;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.zip.Deflater;

// http://www.libpng.org/pub/png/spec/
public class PNG {
	private final static int crcTable[] = new int[256];
	private final static byte ctype[] = {
			-1, 0, 4, 2, 6
	};
	private final static byte sig[] = {
			(byte) 137, 80, 78, 71, 13, 10, 26, 10
	};
	private final static int firstmap[] = {
			0, 1, 2, 3, 4
	};
	private final static int secondmap[] = {
			0, 1, 0, 5, 6
	};
	private static int index = 0;

	private static byte paethFilter(int a, int b, int c) {
		a &= 0xFF;
		b &= 0xFF;
		c &= 0xFF;

		int p = a + b - c, pa = Math.abs(p - a), pb = Math.abs(p - b), pc = Math.abs(p - c);
		if (pa <= pb && pa <= pc) {
			return (byte) a;
		}
		if (pb <= pc) {
			return (byte) b;
		}
		return (byte) c;
	}

	private static void write4bytes(byte[] data, int a, int b, int c, int d) {
		data[index + 0] = (byte) (a & 0xFF);
		data[index + 1] = (byte) (b & 0xFF);
		data[index + 2] = (byte) (c & 0xFF);
		data[index + 3] = (byte) (d & 0xFF);
		index += 4;
	}

	private static void writeInteger(byte[] data, int num) {
		write4bytes(data, num >>> 24, num >>> 16, num >>> 8, num);
	}

	private static void writeTag(byte[] data, String tag) {
		write4bytes(data, tag.charAt(0), tag.charAt(1), tag.charAt(2), tag.charAt(3));
	}

	private static void initCrcTable() {
		int i, j;
		if (crcTable[1] == 0) {
			for (i = 0; i < 256; i++) {
				for (crcTable[i] = i, j = 0; j < 8; ++j) {
					crcTable[i] = (crcTable[i] >>> 1) ^ ((crcTable[i] & 1) != 0 ? 0xEDB88320 : 0);
				}
			}
		}
	}

	private static int crc32_calc(byte[] buffer, int len) {
		int crc = ~0;
		int i;

		initCrcTable();

		for (i = index - len; i < index; i++) {
			crc = (crc >>> 8) ^ crcTable[(buffer[i] ^ crc) & 0xFF];
		}
		return ~crc;
	}

	private static void crc32(byte[] data, int len) {
		int crc = crc32_calc(data, len + 4);
		writeInteger(data, crc);
	}

	public static boolean write(File file, byte[] pixels, int strides, int width, int height, int comp) throws IOException {
		byte filt[];
		byte out[];
		byte line_buffer[];
		int i, j, k, p;

		if (strides == 0) {
			strides = width * comp;
		}

		filt = new byte[(width * comp + 1) * height];
		line_buffer = new byte[width * comp];

		for (j = 0; j < height; j++) {
			int mymap[] = j != 0 ? firstmap : secondmap;
			int best = 0, bestval = 0x7FFFFFFF;
			for (p = 0; p < 2; p++) {
				for (k = (p != 0) ? best : 0; k < 5; k++) {
					int type = mymap[k], est = 0;

					for (i = 0; i < comp; i++) {
						switch (type) {
						case 0:
							line_buffer[i] = pixels[strides * j + i];
							break;
						case 1:
							line_buffer[i] = pixels[strides * j + i];
							break;
						case 2:
							line_buffer[i] = (byte) ((pixels[strides * j + i] & 0xFF - pixels[strides * j + (i - strides)] & 0xFF) & 0xFF);
							break;
						case 3:
							line_buffer[i] = (byte) ((pixels[strides * j + i] & 0xFF - ((pixels[strides * j + i - strides] & 0xFF) >>> 1) & 0xFF) & 0xFF);
							break;
						case 4:
							line_buffer[i] = (byte) (pixels[strides * j + i] & 0xFF - paethFilter(0, pixels[strides * j + i - strides], 0) & 0xFF);
							break;
						case 5:
							line_buffer[i] = pixels[strides * j + i];
							break;
						case 6:
							line_buffer[i] = pixels[strides * j + i];
							break;
						}
					}
					for (i = comp; i < width * comp; ++i) {
						switch (type) {
						case 0:
							line_buffer[i] = pixels[strides * j + i];
							break;
						case 1:
							line_buffer[i] = (byte) (pixels[strides * j + i] - pixels[strides * j + i - comp]);
							break;
						case 2:
							line_buffer[i] = (byte) (pixels[strides * j + i] - pixels[strides * j + i - strides]);
							break;
						case 3:
							line_buffer[i] = (byte) (pixels[strides * j + i] - ((pixels[strides * j + i - comp] + pixels[strides * j + i - strides]) >> 1));
							break;
						case 4:
							line_buffer[i] = (byte) (pixels[strides * j + i] - paethFilter(pixels[strides * j + i - comp], pixels[strides * j + i - strides], pixels[strides * j + i - strides - comp]));
							break;
						case 5:
							line_buffer[i] = (byte) (pixels[strides * j + i] - (pixels[strides * j + i - comp] >> 1));
							break;
						case 6:
							line_buffer[i] = (byte) (pixels[strides * j + i] - paethFilter(pixels[strides * j + i - comp], 0, 0));
							break;
						}
					}
					if (p != 0) {
						break;
					}
					for (i = 0; i < width * comp; ++i) {
						est += Math.abs(line_buffer[i]);
					}
					if (est < bestval) {
						bestval = est;
						best = k;
					}
				}
			}

			filt[j * (width * comp + 1)] = (byte) (best & 0xFF);
			System.arraycopy(line_buffer, 0, filt, j * (width * comp + 1) + 1, width * comp);
		}

		Deflater deflater = new Deflater();
		deflater.setInput(filt);
		deflater.finish();

		ByteArrayOutputStream baos = new ByteArrayOutputStream();
		byte[] buf = new byte[8192];
		while (!deflater.finished()) {
			int byteCount = deflater.deflate(buf);
			baos.write(buf, 0, byteCount);
		}
		deflater.end();

		byte[] zlib = baos.toByteArray();

		out = new byte[sig.length + 12 + 13 + 12 + zlib.length + 12];

		index = 0;
		System.arraycopy(sig, 0, out, 0, sig.length);
		index += sig.length;

		writeInteger(out, 13);
		writeTag(out, "IHDR");
		writeInteger(out, width);
		writeInteger(out, height);

		out[index++] = 8;
		out[index++] = ctype[comp];
		out[index++] = 0;
		out[index++] = 0;
		out[index++] = 0;

		crc32(out, 13);

		writeInteger(out, zlib.length);
		writeTag(out, "IDAT");
		System.arraycopy(zlib, 0, out, index, zlib.length);
		index += zlib.length;
		crc32(out, zlib.length);

		writeInteger(out, 0);
		writeTag(out, "IEND");
		crc32(out, 0);

		FileOutputStream outs = new FileOutputStream(file);
		outs.write(out);

		outs.close();

		return true;
	}
}
