package tjm.tdg.entities;

import java.util.ArrayList;
import java.util.List;

import tjm.tdg.GameWorld;
import tjm.tdg.ResourceManager;
import tjm.tdg.ResourceManager.MODEL;
import tjm.tdg.ResourceManager.SOUND;
import tjm.tdg.entities.ai.AITask;
import tjm.tdg.entities.projectiles.Projectile;
import tjm.tdg.utils.Direction;
import tjm.tdg.utils.Vector2f;

import com.threed.jpct.Object3D;

public abstract class Entity {
	public GameWorld world = null;
	public EntityType type = EntityType.ENTITY;
	public Vector2f position = new Vector2f();
	public Vector2f positionPrev = new Vector2f();
	protected Vector2f positionDiff = new Vector2f();
	public Vector2f radius = new Vector2f();
	public boolean isBlocking = false;
	public int health = 1, maxHealth = 1;
	public boolean isImmortal = false;
	public int attackDamage = 0;
	public double walkSpeed = 0;
	public Direction direction = Direction.NONE;
	public Vector2f directionPrev = this.direction.getVector2f();
	public List<Vector2f> waypoints = new ArrayList<Vector2f>(10);
	public List<AITask> taskList = new ArrayList<AITask>(10);
	public Team team = Team.NONE;

	public float modelOffsetY = 0.0f;

	public Object3D model = ResourceManager.getModelCopy(MODEL.TOWER_BASE);

	@Deprecated
	public boolean isRemoved() {
		return isDead();
	}

	public boolean isDead() {
		return this.health <= 0 && !this.isImmortal;
	}

	public void setDead() {
		this.health = -1;
		this.isImmortal = false;
	}

	public boolean isInRange(Entity entity) {
		return false;
	}

	public boolean isInRange(Entity entity, double attackRange) {
		return getDistance(entity) <= attackRange;
	}

	public List<Entity> entitiesInRange(boolean isSameTeam) {
		List<Entity> list = new ArrayList<Entity>();

		for (Entity entity : this.world.entities) {
			if (this != entity && (this.team == entity.team) == isSameTeam && (entity.type.isMob() || entity.type.isTower())) {
				if (isInRange(entity)) {
					list.add(entity);
				}
			}
		}

		return list;
	}

	public Vector2f getDirectionVector() {
		return this.direction.getVector2f();
	}

	public double getDistance(Entity entity) {
		return this.position.clone().sub(entity.position).length();
	}

	public Entity closestEntity(EntityType[] entityTypes, TriBoolean isSameTeam, TriBoolean isWounded) {
		Entity target = null;
		double distance = 10e10;
		double currDistance = 0;
		for (int i = 0; i < this.world.entities.size(); i++) {
			Entity entity = this.world.entities.get(i);
			currDistance = entity.getDistance(entity);

			if (currDistance >= distance) {
				continue;
			}

			if (skipEntity(this.team == entity.team, isSameTeam)) {
				continue;
			}

			if (skipEntity(entity.health < entity.maxHealth, isWounded)) {
				continue;
			}

			boolean typeMatches = false;
			for (EntityType entityType : entityTypes) {
				if (entity.type == entityType) {
					typeMatches = true;
				}
			}

			if (typeMatches) {
				target = entity;
				distance = currDistance;
			}
		}

		return target;
	}

	private boolean skipEntity(boolean condition, TriBoolean state) {
		switch (state) {
		case NO:
			return condition;
		case YES:
			return !condition;
		case DONT_CARE:
			return false;
		default:
			return false;
		}
	}

	public boolean canAttack() {
		return false;
	}

	public Projectile attack(Entity entity) {
		return null;
	}

	public void update(int ticks) {
		if (isDead()) {
			return;
		}

		this.positionPrev.copyFrom(this.position);
		this.directionPrev = getDirectionVector();

		AITask.executeTask(this, ticks);

		this.positionDiff.copyFrom(this.position).sub(this.positionPrev);

		if (this.positionDiff.length() > 0.0) {
			this.model.translate(this.positionDiff.x, 0, this.positionDiff.y);
		}

		// TODO: rotation diff etc, this will just spin
		// this.model.rotateY(0.1f);
	}

	public String getAttackSound() {
		return SOUND.ATTACK01;
	}
}
